/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.definition.activity;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.FileReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.lang.reflect.Method;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.bonita.connector.Connector;
import org.ow2.bonita.definition.Hook;
import org.ow2.bonita.definition.JavaHook;
import org.ow2.bonita.definition.MultiInstantiator;
import org.ow2.bonita.definition.MultiInstantiatorDescriptor;
import org.ow2.bonita.definition.PerformerAssign;
import org.ow2.bonita.definition.RoleMapper;
import org.ow2.bonita.definition.TxHook;
import org.ow2.bonita.facade.def.element.Resource;
import org.ow2.bonita.facade.def.majorElement.PackageDefinition;
import org.ow2.bonita.facade.exception.BonitaWrapperException;
import org.ow2.bonita.facade.exception.HookInvocationException;
import org.ow2.bonita.facade.runtime.ActivityBody;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.runtime.ClassDataLoader;
import org.ow2.bonita.runtime.XpdlExecution;
import org.ow2.bonita.runtime.XpdlInstance;
import org.ow2.bonita.util.AccessorUtil;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.DateUtil;
import org.ow2.bonita.util.EngineEnvTool;
import org.ow2.bonita.util.Misc;


public class HookExecutorImpl implements HookExecutor {

  private static final Logger LOG = Logger.getLogger(HookExecutorImpl.class.getName());

  public HookExecutorImpl() { }

  /* (non-Javadoc)
   * @see org.ow2.bonita.definition.activity.HookExecutor#
   * executeHook(org.ow2.bonita.runtime.XpdlExecution, java.lang.String, org.ow2.bonita.definition.JavaHook)
   */
  public void executeHook(final XpdlExecution execution, final String activityId, final JavaHook javaHook) {
    if (javaHook != null) {
      final XpdlInstance instance = execution.getXpdlInstance();
      final ProcessInstanceUUID instanceUUID = instance.getUUID();
      final ActivityInstanceUUID activityInstanceUUID = execution.getCurrentActivityInstanceUUID();
      final ProcessDefinitionUUID processUUID = instance.getProcessDefinitionUUID();
      final boolean inTransaction  = javaHook.isInTransaction();
      try {
        final Object hookInstance = ClassDataLoader.getInstance(Hook.class, instance.getPackageDefinitionUUID(), javaHook);
        if (!inTransaction && hookInstance instanceof TxHook) {
          throw new BonitaRuntimeException("The specified hook eventName declares an 'outside transaction hook' "
            + "but the given class implements " + TxHook.class.getName()  + " instead of " + Hook.class.getName());
        }
        if (LOG.isLoggable(Level.FINE)) {
          LOG.fine("Starting hook (instance=" + instanceUUID + ", process=" + processUUID
              + ", activityId=" + activityId + ") : " + javaHook);
        }
        final ActivityInstance<ActivityBody> activityInst =
          EngineEnvTool.getJournalQueriers().getActivityInstance(activityInstanceUUID);
        if (hookInstance instanceof Hook) {
          final Hook hook = (Hook) hookInstance;
          hook.execute(AccessorUtil.getQueryAPIAccessor(), activityInst);
        } else if (hookInstance instanceof TxHook) {
          final TxHook txHook = (TxHook) hookInstance;
          Map<String, String> parameters = javaHook.getParameters();
          if (txHook instanceof Connector) {
            //setParametersFromFile
            if (parameters.containsKey("file:")) {
              BufferedReader in = new BufferedReader(
                  new FileReader(parameters.get("file:")));
              setParametersFromFile(txHook, in);
              parameters.remove("file:");
            } else if (parameters.containsKey("resource:")) {
              InputStream stream =
                ClassDataLoader.getCommonClassLoader().getResourceAsStream(
                    parameters.get("resource:"));
              setParametersFromFile(txHook,
                  new BufferedReader(new InputStreamReader(stream)));
              parameters.remove("resource:");
            } else if (parameters.containsKey("bar:")) {
              PackageDefinitionUUID packageDefUUID =
                  instance.getPackageDefinitionUUID();
              PackageDefinition packageDef =
                  AccessorUtil.getQueryDefinitionAPI().getPackage(packageDefUUID);
              Resource r =
                  packageDef.getBusinessArchive().getResource(parameters.get("bar:"));
              byte[] data = r.getData();
              BufferedReader reader = new BufferedReader(
                  new InputStreamReader(new ByteArrayInputStream(data)));
              setParametersFromFile(txHook, reader);
              parameters.remove("bar:");
            }
            //setInParameters
            for (Entry<String, String> param : parameters.entrySet()) {
              String methodName = param.getKey();
              if (methodName.startsWith("set")) {
                String variableId = param.getValue();
                // check in activity or process variables           
                Object args =
                	AccessorUtil.getQueryRuntimeAPI().getVariable(activityInstanceUUID, variableId);
                Method m =
                  txHook.getClass().getMethod(methodName, args.getClass());
                m.invoke(txHook, args);
              }
            }
          }
          txHook.execute(AccessorUtil.getAPIAccessor(), activityInst);
          if (txHook instanceof Connector) {
            //setOutParameters
            for (Entry<String, String> param : parameters.entrySet()) {
              String methodName = param.getKey();
              if (methodName.startsWith("get")) {
                String variableId = param.getValue();
                Method m =
                  txHook.getClass().getMethod(methodName, new Class[0]);
                Object variableValue = m.invoke(txHook, new Object[0]);
                // check activity or process
                AccessorUtil.getRuntimeAPI().setVariable(activityInstanceUUID, variableId, variableValue);
              }
            }
          }
        } else {
          throw new BonitaRuntimeException("Specified class is not a valid hook. It does not implement "
            + Hook.class.getName() + " neither " + TxHook.class.getName());
        }
        if (LOG.isLoggable(Level.FINE)) {
          LOG.fine("Finished hook (instance=" + instanceUUID + ", process=" + processUUID
              + ", activityId=" + activityId + ") : " + javaHook);
        }
      } catch (final Exception e) {
        if (inTransaction) {
          throw new BonitaWrapperException(
              new HookInvocationException(javaHook.toString(), e)
          );
        } else if (LOG.isLoggable(Level.SEVERE)) {
          LOG.severe("Exception caught while executing hook (instance=" + instanceUUID + ", process=" + processUUID
              + ", activityId=" + activityId + ") : " + javaHook + " - Exception : " + Misc.getStackTraceFrom(e));
        }
      }
    }
  }

  /* (non-Javadoc)
   * @see org.ow2.bonita.definition.activity.HookExecutor#
   * executeMultiInstantiator(org.ow2.bonita.runtime.XpdlExecution, java.lang.String, org.ow2.bonita.definition.MultiInstantiator)
   */
  public MultiInstantiatorDescriptor executeMultiInstantiator(
      final XpdlExecution execution,
      final String activityId,
      final MultiInstantiator actInstantiator) throws Exception {
    return actInstantiator.execute(
        AccessorUtil.getQueryAPIAccessor(),
        execution.getXpdlInstance().getUUID(), activityId, execution.getIterationId());
  }

  /* (non-Javadoc)
   * @see org.ow2.bonita.definition.activity.HookExecutor#
   * executePerformerAssign(org.ow2.bonita.definition.PerformerAssign,
   * org.ow2.bonita.facade.runtime.ActivityInstance, java.util.Set)
   */
  public String executePerformerAssign(
      final PerformerAssign performerAssign,
      final ActivityInstance<ActivityBody> activityInstance,
      final Set<String> candidates) throws Exception {
    return performerAssign.selectUser(AccessorUtil.getQueryAPIAccessor(), activityInstance, candidates);
  }

  /* (non-Javadoc)
   * @see org.ow2.bonita.definition.activity.HookExecutor#
   * executeRoleMapper(org.ow2.bonita.definition.RoleMapper, org.ow2.bonita.facade.uuid.ProcessInstanceUUID, java.lang.String)
   */
  public Set<String> executeRoleMapper(final RoleMapper roleMapper,
      final ProcessInstanceUUID instanceUUID,
      final String roleId) throws Exception {
    return roleMapper.searchMembers(AccessorUtil.getQueryAPIAccessor(), instanceUUID, roleId);
  }
  /* (non-Javadoc)
   * @see org.ow2.bonita.definition.activity.HookExecutor#
   * executeHooks(java.util.List, org.ow2.bonita.runtime.XpdlExecution, java.lang.String, org.ow2.bonita.definition.JavaHook.Type)
   */
  public void executeHooks(final List<JavaHook> javaHooks,
      final XpdlExecution execution, final String activityId, final JavaHook.Type type) {
    if (javaHooks != null) {
      for (final JavaHook javaHook : javaHooks) {
        if (javaHook.getType().equals(type)) {
          executeHook(execution, activityId, javaHook);
        }
      }
    }
  }
  
  private void setParametersFromFile(TxHook txHook, BufferedReader in) throws Exception {
    String line = in.readLine();
    while (line != null) {
      String[] paramter = line.split(" ", 3);
      String setterName = paramter[0].trim();
      String value = paramter[2].trim();
      Object val = null;
      String paramType = paramter[1].trim();
      Class<?> paramClass = null;
      if (paramType.equals("STRING")) {
        paramClass = String.class;
        val = value;
      } else if (paramType.equals("INTEGER")) {
        paramClass = Long.class;
        val = Long.valueOf(value);
      } else if (paramType.equals("FLOAT")) {
        paramClass = Double.class;
        val = Double.valueOf(value);
      } else if (paramType.equals("BOOLEAN")) {
        paramClass = Boolean.class;
        val = Boolean.valueOf(value);
      } else if (paramType.equals("DATETIME")) {
        paramClass = Date.class;
        val = DateUtil.parseDate(value);
      }
      Method m = txHook.getClass().getMethod(setterName, paramClass);
      m.invoke(txHook, val);
      line = in.readLine();
    }
  }
}
